/*
 *  Copyright (C) 2004 by Digital Mars, www.digitalmars.com
 *  Written by Walter Bright
 *
 *  This software is provided 'as-is', without any express or implied
 *  warranty. In no event will the authors be held liable for any damages
 *  arising from the use of this software.
 *
 *  Permission is granted to anyone to use this software for any purpose,
 *  including commercial applications, and to alter it and redistribute it
 *  freely, in both source and binary form, subject to the following
 *  restrictions:
 *
 *  o  The origin of this software must not be misrepresented; you must not
 *     claim that you wrote the original software. If you use this software
 *     in a product, an acknowledgment in the product documentation would be
 *     appreciated but is not required.
 *  o  Altered source versions must be plainly marked as such, and must not
 *     be misrepresented as being the original software.
 *  o  This notice may not be removed or altered from any source
 *     distribution.
 */


module object;

extern (C)
{   int printf(char *, ...);
    int memcmp(void *, void *, size_t);
}


alias bit bool;

version (AMD64)
{
    alias ulong size_t;
    alias long ptrdiff_t;
}
else
{
    alias uint size_t;
    alias int ptrdiff_t;
}

class Object
{
    void print()
    {
  printf("%.*s\n", toString());
    }

    char[] toString()
    {
  return this.classinfo.name;
    }

    uint toHash()
    {
  // BUG: this prevents a compacting GC from working, needs to be fixed
  return cast(uint)cast(void *)this;
    }

    int opCmp(Object o)
    {
  // BUG: this prevents a compacting GC from working, needs to be fixed
  return cast(int)cast(void *)this - cast(int)cast(void *)o;
    }

    int opEquals(Object o)
    {
  return this is o;
    }
}

struct Interface
{
    ClassInfo classinfo;
    void *[] vtbl;
    int offset;     // offset to Interface 'this' from Object 'this'
}

class ClassInfo : Object
{
    byte[] init;    // class static initializer
    char[] name;    // class name
    void *[] vtbl;    // virtual function pointer table
    Interface[] interfaces;
    ClassInfo base;
    void *destructor;
    void (*classInvariant)(Object);
    uint flags;
    //  1:      // IUnknown
    void *deallocator;
}

private import std.string;

class TypeInfo
{
    uint toHash()
    { uint hash;

  foreach (char c; this.classinfo.name)
      hash = hash * 9 + c;
  return hash;
    }

    int opCmp(Object o)
    {
  return std.string.cmp(this.classinfo.name, o.classinfo.name);
    }

    int opEquals(Object o)
    {
  /* TypeInfo instances are singletons, but duplicates can exist
   * across DLL's. Therefore, comparing for a name match is
   * sufficient.
   */
  return this is o || this.classinfo.name == o.classinfo.name;
    }

    uint getHash(void *p) { return cast(uint)p; }
    int equals(void *p1, void *p2) { return p1 == p2; }
    int compare(void *p1, void *p2) { return 0; }
    int tsize() { return 0; }
    void swap(void *p1, void *p2)
    {
  int i;
  int n = tsize();
  for (i = 0; i < n; i++)
  {   byte t;

      t = (cast(byte *)p1)[i];
      (cast(byte *)p1)[i] = (cast(byte *)p2)[i];
      (cast(byte *)p2)[i] = t;
  }
    }
}

class TypeInfo_Typedef : TypeInfo
{
    uint getHash(void *p) { return base.getHash(p); }
    int equals(void *p1, void *p2) { return base.equals(p1, p2); }
    int compare(void *p1, void *p2) { return base.compare(p1, p2); }
    int tsize() { return base.tsize(); }
    void swap(void *p1, void *p2) { return base.swap(p1, p2); }

    TypeInfo base;
}

class TypeInfo_Class : TypeInfo
{
    char[] toString() { return info.name; }

    uint getHash(void *p)
    {
  Object o = *cast(Object*)p;
  assert(o);
  return o.toHash();
    }

    int equals(void *p1, void *p2)
    {
  Object o1 = *cast(Object*)p1;
  Object o2 = *cast(Object*)p2;

  return o1 == o2 || (o1 && o1.opCmp(o2) == 0);
    }

    int compare(void *p1, void *p2)
    {
  Object o1 = *cast(Object*)p1;
  Object o2 = *cast(Object*)p2;
  int c = 0;

  // Regard null references as always being "less than"
  if (o1 != o2)
  {
      if (o1)
      { if (!o2)
        c = 1;
    else
        c = o1.opCmp(o2);
      }
      else
    c = -1;
  }
  return c;
    }

    int tsize()
    {
  return Object.sizeof;
    }

    ClassInfo info;
}

class TypeInfo_Struct : TypeInfo
{
    char[] toString() { return name; }

    uint getHash(void *p)
    { uint h;

  assert(p);
  if (xtoHash)
  {   //printf("getHash() using xtoHash\n");
      h = (*xtoHash)(p);
  }
  else
  {
      //printf("getHash() using default hash\n");
      // A sorry hash algorithm.
      // Should use the one for strings.
      // BUG: relies on the GC not moving objects
      for (size_t i = 0; i < xsize; i++)
      { h = h * 9 + *cast(ubyte*)p;
    p++;
      }
  }
  return h;
    }

    int equals(void *p2, void *p1)
    { int c;

  if (p1 == p2)
      c = 1;
  else if (!p1 || !p2)
      c = 0;
  else if (xopEquals)
      c = (*xopEquals)(p1, p2);
  else
      // BUG: relies on the GC not moving objects
      c = (memcmp(p1, p2, xsize) == 0);
  return c;
    }

    int compare(void *p2, void *p1)
    {
  int c = 0;

  // Regard null references as always being "less than"
  if (p1 != p2)
  {
      if (p1)
      { if (!p2)
        c = 1;
    else if (xopCmp)
        c = (*xopCmp)(p1, p2);
    else
        // BUG: relies on the GC not moving objects
        c = memcmp(p1, p2, xsize);
      }
      else
    c = -1;
  }
  return c;
    }

    int tsize()
    {
  return xsize;
    }

    char[] name;
    uint xsize;

    uint function(void*) xtoHash;
    int function(void*,void*) xopEquals;
    int function(void*,void*) xopCmp;
}




/*

class Exception : Object
{
  
  DWORD abc;
  
    char[] msg;

    this(char[] msg)
    {
version (Windows) 
{
      DWORD params[1];
      params[0] = cast(DWORD)&abc;
      
      printf("abc=%d", abc);
//      RaiseException(0x0e0440002, 0, 1, params);
      fill_stack_trace();
      printf("abc=%d", abc);
}

  this.msg = msg;
    }

    void print()
    {
  printf("%.*s\n", msg);
    }

    char[] toString() { return msg; }
}

class Error : Exception
{
    Error next;

    this(char[] msg)
    {
  super(msg);
    }

    this(char[] msg, Error next)
    {
  super(msg);
  this.next = next;
    }
}

//extern (C) int nullext = 0;
*/


















//
// Copyright (c) 2005
// All Rights Reserved
// Written by Maxime Larose


private import std.string;

version (Windows) {
  private  {
    import std.c.windows.windows;
    import std.c.windows.dbghelp;
  }
}

 
class Exception : Object
{
  char[]     msg;
  Exception  cause;
  StackTrace stackTrace;
 
  // used internally
  private this()
  {
    this.msg = null;
    this.cause = null;
    this.stackTrace = StackTrace.newStackTrace();
  }
  this(char[] msg)
  {
    // force stack frame
    asm { nop; }
    this.msg = msg;
    this.cause = null;
    this.stackTrace = StackTrace.newStackTrace();
  }
  this(char[] msg, Exception cause)
  {
    // force stack frame
    asm { nop; }
    this.msg = msg;
    this.cause = cause;
    this.stackTrace = StackTrace.newStackTrace();
  }



  void print()
  {
    printf("%.*s", toString());
  }

  char[] toString() 
  {  
    char[] str;
    str ~= this.classinfo.name;
    str ~= ":\"";
    str ~= msg;
    str ~= "\"\n";
    // should never be null...
    if (!(stackTrace is null)) {
      str ~= stackTrace.toString();
    } else {
      str ~= "  no stack trace: internal error\n";
    }

    
    if (!(cause is null)) {
      str ~= "Caused by:\n";
      str ~= cause.toString();
    }

    return str;
  }
}



class AccessViolationException : Exception
{
  this(char[] msg)
  {
    // force stack frame
    asm { nop; }
    super(msg);
  }

  this(char[] msg, Exception cause)
  {
    // force stack frame
    asm { nop; }
    super(msg, cause);
  }
}


class DivisionByZeroException : Exception
{
  this(char[] msg)
  {
    // force stack frame
    asm { nop; }
    super(msg);
  }

  this(char[] msg, Exception cause)
  {
    // force stack frame
    asm { nop; }
    super(msg, cause);
  }
}


class StackOverflowException : Exception
{
  this(char[] msg)
  {
    // force stack frame
    asm { nop; }
    super(msg);
  }

  this(char[] msg, Exception cause)
  {
    // force stack frame
    asm { nop; }
    super(msg, cause);
  }
}


class Error : Exception
{
  this(char[] msg)
  {
    // force stack frame
    asm { nop; }
    super(msg);
  }

  this(char[] msg, Exception cause)
  {
    // force stack frame
    asm { nop; }
    super(msg, cause);
  }
}


 

private class StackTrace : Object
{

  version (Windows) {
    version (X86) {
      alias fillStackTrace_windows_x86 fillStackTrace;
    }
    else {
      alias fillStackTrace_not_supported fillStackTrace;
    }
  }
  else {
    alias fillStackTrace_not_supported fillStackTrace;
  }


  static Object lock;
  static HANDLE hProcess;
  static bit symbolsLoaded;
  static DWORD reasonSymbolsNotLoaded;


  const int MAX_STACK_TRACE_DEPTH = 25;
  char[][MAX_STACK_TRACE_DEPTH] stackTrace;
  int nbElements;


  static this()
  {
    lock = new Object();
    hProcess = GetCurrentProcess();
    SymSetOptions(SYMOPT_LOAD_LINES | SYMOPT_DEFERRED_LOADS);

    symbolsLoaded = true;
    reasonSymbolsNotLoaded = 0;
    if (!SymInitialize( hProcess, null, TRUE )) {
      symbolsLoaded = false;
      reasonSymbolsNotLoaded = GetLastError();
    }
  }

  static ~this()
  {
    SymCleanup(hProcess);
  }

  // Here, we force a stack frame (even in release mode)
  // So, we can later check for the Exception ctor and
  // start the stack trace from there
  static StackTrace newStackTrace()
  {
    asm { naked; 
          enter 0,0; 
    }
    new StackTrace();
    // EAX contains the result
    asm { leave; 
          ret;
    }
  }


  private this()
  {
    fillStackTrace();
  }


  char[] toString()
  { 
    char[] str = "";
    for (int i; i < nbElements; ++i) {
      str ~= stackTrace[i] ~ "\n";
    }

    return str;
  }


  char[] unmangle(char[] symbol)
  {
    char[] unmangled;
    char[] part;

    if (find(symbol, "__Dmain") == 0) {
      return "main";
    }

    // if it's not a D name, forget it
    if (find(symbol, "_D") != 0) {
      return symbol;
    }

    int charAt = 2;
    int len;

    while (true) {
      if (symbol[charAt] == 'F') {
        break;
      }

      unmangled ~= ".";
      len = atoi(symbol[charAt..length]);
      if (len == 0) {
        break;
      }
      ++charAt;
      if (len > 9) {
        ++charAt;
      }

      part = symbol[charAt..charAt+len];
      if (find(part, "_ctor") == 0) {
        part = "this";
      }

      unmangled ~= part;
      charAt += len;
    }

    return unmangled;
  }


  void fillStackTrace_not_supported()
  {
    stackTrace[0] = "  Stack trace unavailable: not implemented";
    nbElements = 1;
  }



  version (Windows) 
  {
    version (X86) {
      /* Having a CONTEXT in the next function would be good but I 
      don't know how to code an __except with DM C++...
      try {
        RaiseException(1, 0, 0, NULL);
      }
      __except(setExceptionPointers(GetExceptionInformation()) {
      }
      */
      void fillStackTrace_windows_x86()
      {
        nbElements = 0;

        if (!symbolsLoaded) {
          stackTrace[nbElements] = "  Stack trace unavailable: cannot load symbols (err=";
          stackTrace[nbElements] ~= .toString(reasonSymbolsNotLoaded);
          stackTrace[nbElements] ~= ")";
          ++nbElements;
        } 
        else {
          STACKFRAME64 sf;
          memset( &sf, 0, sf.sizeof );

          // put some registers in the stack frame struct
          // (we should instead have a CONTEXT...)
          DWORD eipValue;
          DWORD espValue;
          DWORD ebpValue;

          // getting EIP is tricky
          asm { 
                  call get_eip;
            get_eip:
                  pop  EAX;

                  mov eipValue,EAX;
                  mov espValue,ESP;
                  mov ebpValue,EBP;
          }

          sf.AddrPC.Offset       = eipValue;
          sf.AddrPC.Mode         = ADDRESS_MODE.AddrModeFlat;
          sf.AddrStack.Offset    = espValue;
          sf.AddrStack.Mode      = ADDRESS_MODE.AddrModeFlat;
          sf.AddrFrame.Offset    = ebpValue;
          sf.AddrFrame.Mode      = ADDRESS_MODE.AddrModeFlat;


          // prepare some struct
          IMAGEHLP_LINE64 line;
          line.SizeOfStruct = line.sizeof;
          DWORD64 symDisplacement = 0;

          byte buffer[IMAGEHLP_SYMBOL64.sizeof + 1024 ];
          PIMAGEHLP_SYMBOL64 pSymbol = cast(PIMAGEHLP_SYMBOL64)&buffer;
          pSymbol.SizeOfStruct = buffer.sizeof;
          pSymbol.MaxNameLength = 1024;
          DWORD lineDisplacement = 0;  

          char[] symbolName;
          HANDLE hThread = GetCurrentThread();

          synchronized(lock) {
            while ( nbElements < MAX_STACK_TRACE_DEPTH )
            {
              if ( ! StackWalk64(IMAGE_FILE_MACHINE_I386,
                                 hProcess,
                                 hThread,
                                 &sf,
                                 null,
                                 null,
                                 &SymFunctionTableAccess64,
                                 &SymGetModuleBase64,
                                 null ) )
                 break;

              if ( 0 == sf.AddrFrame.Offset ) // check for null
                 break;

		          // get symbol
		          if (SymGetSymFromAddr64(hProcess, sf.AddrPC.Offset,
								         &symDisplacement, pSymbol) )
		          {

                symbolName = .toString(&pSymbol.Name);

                if (find(symbolName, "Exception5_ctor") != -1) {
                    // we have found an exception constructor frame
                    // trash what we had and start from here
                    nbElements = 0;
                    continue;
                }

                // this is Windows entry point when processing 
                // an exception
                // hardcoding is ugly, but I don't believe 
                // there is any other way...
                // at worst, the string will not be found 
                // and we will end up with a more complete (and ugly) 
                // stack trace
                if (find(symbolName, "RtlConvertUiListToApiList") == 0) {
                    // we have found the exception entry point
                    // trash what we had and start from here
                    nbElements = 0;
                    continue;
                }

                if (find(symbolName, "_mainCRTStartup") == 0) {
                  // we have enough
                  break;
                }

                stackTrace[nbElements] = "  at ";
                stackTrace[nbElements] ~= unmangle(symbolName);

                // substract because we want the offending address, 
                // not the address of the return
                DWORD64 addr = sf.AddrPC.Offset - 2;

                // get line
			          if ( SymGetLineFromAddr64(hProcess, addr,
									         &lineDisplacement, &line) )
			          {
                  stackTrace[nbElements] ~= " (";
                  stackTrace[nbElements] ~= .toString(line.FileName);
                  stackTrace[nbElements] ~= ":";
                  stackTrace[nbElements] ~= .toString(line.LineNumber);
                  stackTrace[nbElements] ~= ")";
			          }
			          else {
                  stackTrace[nbElements] ~= format("+%04X", symDisplacement);
			          }
                ++nbElements;
		          }
              else {
                stackTrace[nbElements] = "  Stack symbol not found (err=";
                stackTrace[nbElements] ~= .toString(GetLastError());
                stackTrace[nbElements] ~= ")";
                ++nbElements;
              }
            } // while there are stack frames
          } // synchronized lock
        } // if symbols loaded
      } // method
    } // versions X86
  } // version Windows





}






